from typing import Optional, Dict, Any, Type, TYPE_CHECKING
from collections import OrderedDict
import inspect
import copyreg

from .. import component as comp

if TYPE_CHECKING:
    from .typing import PreElabRDLType

UserStructMembers = Dict[str, 'PreElabRDLType']


class UserStructMeta(type):
    """
    Declare a metaclass for UserStruct so that it can be uniquely identified
    during dynamic type pickling
    """
    _members = OrderedDict() # type: UserStructMembers
    _is_abstract = True # type: bool
    _parent_scope = None # type: Optional[comp.Component]

    def define_new(cls, name: str, members: UserStructMembers, is_abstract: bool=False, _parent_scope: Optional[comp.Component]=None) -> Type['UserStruct']:
        """
        Define a new struct type derived from the current type.

        Parameters
        ----------
        name: str
            Name of the struct type
        members: {member_name : type}
            Dictionary of struct member types.
        is_abstract: bool
            If set, marks the struct as abstract.
        """

        # Extend the base struct's members
        m = OrderedDict(cls._members)
        # Make sure derivation does not have any overlapping keys with its parent
        if set(m.keys()) & set(members.keys()):
            raise ValueError("'members' contains keys that overlap with parent")
        m.update(members)

        # Create the new class
        classdict = {
            '_members' : m,
            '_is_abstract': is_abstract,
            '_parent_scope': _parent_scope,
        }
        metaclass = cls.__class__
        return metaclass(name, (cls,), classdict) # type: ignore

    @property
    def type_name(cls) -> str:
        # (docstring is in rst)
        return cls.__name__


class UserStruct(metaclass=UserStructMeta):
    """
    All user-defined struct types are extended from this class and can be
    identified using ``issubclass(obj, UserStruct)``. Once assigned a value,
    the class is constructed and is identified using ``isinstance(obj, UserStruct)``.

    Values of struct members are accessed as read-only object attributes.

    For example, the following RDL struct literal:

    .. code-block:: systemrdl

        struct my_struct {
            longint foo;
            longint bar;
        };
        ...
        my_struct_prop = my_struct'{foo:42, bar:123};

    ... can be queried in Python as follows:

    .. code-block:: python

        prop = node.get_property('my_struct_prop')

        foo = prop.foo
        bar = getattr(prop, "bar")

    Or UserStruct members names can be accessed as a mapping:

    .. code-block:: python

        for member_name, value in prop.members.items():
            ...
    """

    def __init__(self, values: Dict[str, Any]):
        """
        Create an instance of the struct

        values is a dictionary of {member_name : value}
        """
        if self._is_abstract:
            raise TypeError("Cannot create instance of an abstract struct type")

        # Make sure values dict matches the members allowed
        if set(values.keys()) != set(self._members.keys()):
            raise ValueError("Cannot map 'values' to this struct")

        self._values = values


    def __getattr__(self, name: str) -> Any:
        if name == "__setstate__":
            raise AttributeError(name)
        if name == "type_name":
            # getattr interferes with class property mechanism
            return type(self).type_name
        if name in self._values:
            return self._values[name]
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.type_name, name))


    @property
    def members(self) -> Dict[str, Any]:
        """
        Get a dictionary of struct members


        .. versionadded:: 1.24
        """
        return self._values

    @classmethod
    def get_parent_scope(cls) -> Optional[comp.Component]:
        """
        Returns reference to parent component that contains this type definition.
        """
        return getattr(cls, "_parent_scope", None)

    @classmethod
    def get_scope_path(cls, scope_separator: str="::") -> str:
        """
        Generate a string that represents this enum's declaration namespace
        scope.

        Parameters
        ----------
        scope_separator: str
            Override the separator between namespace scopes
        """
        parent_scope = cls.get_parent_scope()
        if parent_scope is None:
            # Importer likely never set the scope
            return ""
        elif isinstance(parent_scope, comp.Root):
            # Declaration was in root scope
            return ""
        else:
            # Get parent definition's scope path
            parent_path = parent_scope.get_scope_path(scope_separator)

            # Extend it with its scope name
            if parent_path:
                return(
                    parent_path
                    + scope_separator
                    + parent_scope._scope_name
                )
            else:
                return parent_scope._scope_name

    def __repr__(self) -> str:
        return "<struct '%s' %s at 0x%x>" % (
            self.__class__.__qualname__,
            "(%s)" % ", ".join(self._members.keys()),
            id(self)
        )


# Tell pickle how to reduce dynamically generated UserStruct classes
def _reduce_user_struct(c: Type[UserStruct]) -> Any:
    if c is UserStruct:
        # Reached base class. Return string so pickle can look up the actual object
        return 'UserStruct'

    assert len(c.__bases__) == 1 # Only supporting single-inheritence
    base_cls = c.__bases__[0] # type: Type[UserStruct]

    # remove members that exist in base class
    unique_members = c._members.copy()
    for k in base_cls._members.keys():
        del unique_members[k]

    args = (c.type_name, unique_members, c._is_abstract, c._parent_scope)
    return (base_cls.define_new, args)
copyreg.pickle(UserStructMeta, _reduce_user_struct) # type: ignore


# Utility functions
def is_user_struct(t: Any) -> bool:
    """
    Test if type ``t`` is a :class:`~UserStruct`
    """
    return inspect.isclass(t) and issubclass(t, UserStruct)



import os
import platform
import setuptools
from glob import glob

target = platform.system().lower()
PLATFORMS = {'windows', 'linux', 'darwin', 'cygwin'}
for known in PLATFORMS:
    if target.startswith(known):
        target = known

def run_setup(with_binary):
    if with_binary:
        extra_compile_args = {
            'windows': ['/DANTLR4CPP_STATIC', '/Zc:__cplusplus', '/std:c++17'],
            'linux': ['-std=c++17'],
            'darwin': ['-std=c++17'],
            'cygwin': ['-std=c++17'],
        }

        # Define an Extension object that describes the Antlr accelerator
        parser_ext = setuptools.Extension(
            name='systemrdl.parser.sa_systemrdl_cpp_parser',

            # Add the Antlr runtime source directory to the include search path
            include_dirs=["src/systemrdl/parser/ext/antlr4-cpp-runtime"],

            # Rather than listing each C++ file (Antlr has a lot!), discover them automatically
            sources=glob("src/systemrdl/parser/ext/**/*.cpp", recursive=True),
            depends=glob("src/systemrdl/parser/ext/**/*.h", recursive=True),

            extra_compile_args=extra_compile_args.get(target, []),
            define_macros=[("Py_LIMITED_API", "0x03070000")],
            py_limited_api=True,
        )
        ext_modules = [parser_ext]
        options = {"bdist_wheel": {"py_limited_api": "cp37"}}
    else:
        ext_modules = []
        options = {}

    setuptools.setup(
        ext_modules=ext_modules,
        cmdclass={"build_ext": ve_build_ext},
        options=options,
    )


#===============================================================================
from setuptools.command.build_ext import build_ext

class BuildFailed(Exception):
    pass

class ve_build_ext(build_ext):
    """
    This class extends setuptools to fail with a common BuildFailed exception
    if a build fails
    """
    def run(self):
        try:
            build_ext.run(self)
        except Exception:
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except Exception:
            raise BuildFailed()

using_fallback = False
try:
    run_setup(with_binary=True)
except BuildFailed:
    if 'SYSTEMRDL_REQUIRE_BINARY_BUILD' in os.environ:
        # Force failure if binary build is required
        raise
    else:
        using_fallback = True

if using_fallback:
    run_setup(with_binary=False)



from typing import Set, Type, Any, List, Dict, Optional, Iterable, TYPE_CHECKING
import warnings as py_warnings

from antlr4 import InputStream

from . import messages
from . import warnings # pylint: disable=reimported
from .parser import sa_systemrdl
from .core.ComponentVisitor import RootVisitor
from .core.ExprVisitor import ExprVisitor
from .properties.rulebook import PropertyRuleBook
from .properties.user_defined import ExternalUserProperty, LegacyExternalUserProperty
from .core.namespace import NamespaceRegistry
from .core.elaborate import ElabExpressionsListener, PrePlacementValidateListener, LateElabListener
from .core.elaborate import StructuralPlacementListener, LateElabRevisitor
from .core.validate import ValidateListener
from . import ast
from . import component as comp
from . import walker
from .node import RootNode
from . import preprocessor

if TYPE_CHECKING:
    from .rdltypes.typing import RDLValue
    from .udp import UDPDefinition


class FileInfo:
    def __init__(self, preprocessed_text: str, included_files: Iterable[str]) -> None:
        self._pp_text = preprocessed_text
        self._incl_files = included_files

    @property
    def preprocessed_text(self) -> str:
        """
        Resolved text after Perl and Verilog preprocessing
        """
        return self._pp_text

    @property
    def included_files(self) -> Iterable[str]:
        """
        Iterable of paths that were included while preprocessing this file.
        """
        return self._incl_files


class RDLCompiler:

    def __init__(self, **kwargs: Any):
        """
        RDLCompiler constructor.

        Parameters
        ----------
        message_printer: :class:`~systemrdl.messages.MessagePrinter`
            Override the default message printer
        warning_flags: int
            Flags to enable warnings. See :ref:`messages_warnings` for more details.
        error_flags: int
            Same as ``warning_flags`` but promote them to errors instead.
        dedent_desc: bool
            Automatically remove any common indentation from multi-line
            ``desc`` properties.

            Set to True by default.
        extended_dpa_type_names: bool
            Enable extended type name generation that accounts for dynamic
            property assignments augmenting the type.

            Set to True by default.

            See :ref:`dpa_type_generation` for more details.
        perl_safe_opcodes: List[str]
            Perl preprocessor commands are executed within a
            `Perl Safe <https://perldoc.perl.org/Safe.html>`_ compartment to
            prevent malicious code execution.

            The default set of `Perl opcodes <https://perldoc.perl.org/Opcode.html#Predefined-Opcode-Tags>`_
            allowed should be sufficient for most applications, however this
            option is exposed in the rare case it is necessary to override the
            opcode list in order to make an exception.

            Default value::

                [
                    ':base_core', ':base_mem', ':base_loop', ':base_orig', ':base_math',
                    ':base_thread', ':filesys_read', ':sys_db', ':load',
                    'sort', 'tied', 'pack', 'unpack', 'reset'
                ]


        .. versionchanged:: 1.8
            Added ``dedent_desc`` option.
        .. versionchanged:: 1.9
            Added ``extended_dpa_type_names`` option.
        .. versionchanged:: 1.10
            Added ``perl_safe_opcodes`` option.
        """
        self.env = RDLEnvironment(kwargs)

        # Check for stray kwargs
        if kwargs:
            raise TypeError(f"got an unexpected keyword argument '{list(kwargs.keys())[0]}'")

        #: Reference to the compiler's :class:`~systemrdl.messages.MessageHandler` object
        #:
        #: .. warning::
        #:
        #:      This will be deprecated in a future release. See this page for more details: https://github.com/SystemRDL/systemrdl-compiler/issues/168
        self.msg = self.env.msg

        self.namespace = NamespaceRegistry(self.env) # type: NamespaceRegistry
        self.visitor = RootVisitor(self)
        self.root = self.visitor.component # type: comp.Root # type: ignore


    def define_udp(
            self, name: str, valid_type: Any,
            valid_components: 'Optional[Set[Type[comp.Component]]]'=None,
            default: Any=None
        ) -> None:

        py_warnings.warn(
            "Use of RDLCompiler.define_udp() is deprecated. Use RDLCompiler.register_udp() instead.",
            DeprecationWarning, stacklevel=2
        )

        if name in self.env.property_rules.rdl_properties:
            raise ValueError(f"UDP definition's name '{name}' conflicts with existing built-in RDL property")
        if name in self.env.property_rules.user_properties:
            raise ValueError(f"UDP '{name}' has already been defined")

        udp = LegacyExternalUserProperty(
            self.env,
            name,
            valid_components,
            valid_type,
            default_assignment=default,
            constr_componentwidth=False
        )
        self.env.property_rules.user_properties[udp.name] = udp

    def register_udp(self, definition_cls: 'Type[UDPDefinition]', soft: bool=True) -> None:
        """
        Pre-register a User Defined Property into the compiler.

        Parameters
        ----------
        definition_cls: :class:`systemrdl.udp.UDPDefinition`
            Reference to the container class that defines your new UDP.
        soft: bool
            Override to False to register the UDP as a hard definition.


        .. versionadded:: 1.25
        """
        if definition_cls.name in self.env.property_rules.rdl_properties:
            raise ValueError(f"UDP definition's name '{definition_cls.name}' conflicts with existing built-in RDL property")
        if definition_cls.name in self.env.property_rules.user_properties:
            raise ValueError(f"UDP '{definition_cls.name}' has already been defined")

        # Wrap definition with internal UDP object & register it
        udp = ExternalUserProperty(self.env, definition_cls, soft)
        self.env.property_rules.user_properties[udp.name] = udp

    def list_udps(self) -> List[str]:
        """
        List all user-defined properties encountered by the compiler.


        .. versionadded:: 1.12
        """
        udps = []
        for udp_name, udp in self.env.property_rules.user_properties.items():
            if isinstance(udp, ExternalUserProperty) and udp.is_soft:
                continue
            udps.append(udp_name)
        return udps


    def preprocess_file(self, path: str, incl_search_paths: Optional[List[str]]=None, defines: Optional[Dict[str,str]]=None) -> FileInfo:
        """
        Preprocess a single file without compiling it.

        Parameters
        ----------
        path:str
            Path to an RDL source file

        incl_search_paths: List[str]
            List of additional paths to search to resolve includes.
            If unset, defaults to an empty list.

            Relative include paths are resolved in the following order:

            1. Search each path specified in ``incl_search_paths``.
            2. Path relative to the source file performing the include.

        defines: Dict[str, str]
            Dictionary of pre-defined verilog macros where the key is the macro
            name, and the value is the macro text.

        Raises
        ------
        RDLCompileError
            If any fatal preprocessing error is encountered.

        Returns
        -------
        :class:`FileInfo`
            File info object


        .. versionadded:: 1.20
        """
        if incl_search_paths is None:
            incl_search_paths = []

        input_stream, included_files = preprocessor.preprocess_file(self.env, path, incl_search_paths, defines)

        return FileInfo(input_stream.strdata, included_files)


    def compile_file(self, path: str, incl_search_paths: Optional[List[str]]=None, defines: Optional[Dict[str,str]]=None) -> FileInfo:
        """
        Parse & compile a single file and append it to RDLCompiler's root
        namespace.

        If any exceptions (:class:`~systemrdl.RDLCompileError` or other)
        occur during compilation, then the RDLCompiler object should be discarded.

        Parameters
        ----------
        path:str
            Path to an RDL source file

        incl_search_paths: List[str]
            List of additional paths to search to resolve includes.
            If unset, defaults to an empty list.

            Relative include paths are resolved in the following order:

            1. Search each path specified in ``incl_search_paths``.
            2. Path relative to the source file performing the include.

        defines: Dict[str, str]
            Dictionary of pre-defined verilog macros where the key is the macro
            name, and the value is the macro text.

        Raises
        ------
        RDLCompileError
            If any fatal compile error is encountered.

        Returns
        -------
        :class:`FileInfo`
            File info object


        .. versionchanged:: 1.20
            Returns a :class:`FileInfo` object instead of ``None``
        """

        if incl_search_paths is None:
            incl_search_paths = []

        input_stream, included_files = preprocessor.preprocess_file(self.env, path, incl_search_paths, defines)

        # Run Antlr parser on input
        parsed_tree = sa_systemrdl.parse(
            input_stream,
            "root",
            messages.RdlSaErrorListener(self.msg)
        )

        if self.msg.had_error:
            self.msg.fatal("Parse aborted due to previous errors")

        # Traverse parse tree with RootVisitor
        self.visitor.visit(parsed_tree)

        # Reset default property assignments from namespace.
        # They should not be shared between files since that would be confusing.
        self.namespace.default_property_ns_stack = [{}]

        if self.msg.had_error:
            self.msg.fatal("Compile aborted due to previous errors")

        return FileInfo(input_stream.strdata, included_files)


    def elaborate(self, top_def_name: Optional[str]=None, inst_name: Optional[str]=None, parameters: Optional[Dict[str, 'RDLValue']]=None) -> RootNode:
        """
        Elaborates the design for the given top-level addrmap component.

        During elaboration, the following occurs:

        - An instance of the ``$root`` meta-component is created.
        - The addrmap component specified by ``top_def_name`` is instantiated as a
          child of ``$root``.
        - Expressions, parameters, and inferred address/field placements are elaborated.
        - Validation checks are performed.

        If a design contains multiple root-level addrmaps, ``elaborate()`` can be
        called multiple times in order to elaborate each individually.

        If any exceptions (:class:`~systemrdl.RDLCompileError` or other)
        occur during elaboration, then the RDLCompiler object should be discarded.

        Parameters
        ----------
        top_def_name: str
            Explicitly choose which addrmap  in the root namespace will be the
            top-level component.

            If unset, The last addrmap defined will be chosen.

        inst_name: str
            Overrides the top-component's instantiated name.
            By default, instantiated name is the same as ``top_def_name``

        parameters: dict
            Dictionary of parameter overrides for the top component instance.

        Raises
        ------
        RDLCompileError
            If any fatal elaboration error is encountered

        Returns
        -------
        :class:`~systemrdl.node.RootNode`
            Elaborated root meta-component's Node object.
        """
        if parameters is None:
            parameters = {}

        # Get top-level component definition to elaborate
        if top_def_name is not None:
            # Lookup top_def_name
            if top_def_name not in self.root.comp_defs:
                self.msg.fatal(f"Elaboration target '{top_def_name}' not found")
            top_def = self.root.comp_defs[top_def_name]

            if not isinstance(top_def, comp.Addrmap):
                self.msg.fatal(f"Elaboration target '{top_def_name}' is not an 'addrmap' component")
        else:
            # Not specified. Find the last addrmap defined
            for comp_def in reversed(self.root.comp_defs.values()):
                if isinstance(comp_def, comp.Addrmap):
                    top_def = comp_def
                    top_def_name = comp_def.type_name
                    break
            else:
                self.msg.fatal("Could not find any 'addrmap' components to elaborate")

        # Create an instance of the root component
        root_inst = self.root._copy_for_inst({})
        root_inst.is_instance = True
        root_inst.original_def = self.root
        root_inst.inst_name = "$root"

        # Create a top-level instance
        top_inst = top_def._copy_for_inst({})
        top_inst.is_instance = True
        top_inst.original_def = top_def
        top_inst.addr_offset = 0
        top_inst.external = True # addrmap is always implied as external
        if inst_name is not None:
            top_inst.inst_name = inst_name
        else:
            top_inst.inst_name = top_def_name

        # Override parameters as needed
        for param_name, value in parameters.items():
            # Find the parameter to override
            parameter = None
            for p in top_inst.parameters:
                if p.name == param_name:
                    parameter = p
                    break
            else:
                self.msg.fatal(f"Elaboration top does not have a parameter '{param_name}' that is available for override")

            literal_expr = ast.ExternalLiteral(self.env, value)
            assign_expr = ast.AssignmentCast(self.env, None, literal_expr, parameter.param_type)
            assign_type = assign_expr.predict_type()
            if assign_type is None:
                self.msg.fatal(f"Incorrect type for top-level parameter override of '{param_name}'")

            parameter.expr = assign_expr


        # instantiate top_inst into the root component instance
        root_inst.children.append(top_inst)

        root_node = RootNode(root_inst, self.env, None)

        # Resolve all expressions
        walker.RDLWalker(skip_not_present=False).walk(
            root_node,
            ElabExpressionsListener(self.msg)
        )

        # Resolve address and field placement
        late_elab_listener = LateElabListener(self.msg, self.env)
        walker.RDLWalker(skip_not_present=False).walk(
            root_node,
            PrePlacementValidateListener(self.msg),
            StructuralPlacementListener(self.msg),
            late_elab_listener
        )

        # re-visit nodes a 2nd time as-needed to complete elaboration
        LateElabRevisitor(late_elab_listener.node_needs_revisit)

        # Validate design
        # Only need to validate nodes that are present
        walker.RDLWalker(skip_not_present=True).walk(root_node, ValidateListener(self.env))

        if self.msg.had_error:
            self.msg.fatal("Elaborate aborted due to previous errors")

        return root_node


    def eval(self, expression: str) ->'RDLValue':
        """
        Evaluate an RDL expression string and return its compiled value.
        This function is provided as a helper to simplify overriding top-level
        parameters during elaboration.

        Parameters
        ----------
        expression: str
            This string is parsed and evaluated as a SystemRDL expression.
            Any references used in the expression are resolved using the
            current contents of the root namespace.

        Raises
        ------
        ValueError
            If any parse or evaluation error occurs.


        .. versionadded:: 1.8
        """
        # Create local message handler that suppresses the usual output
        # to stderr.
        # Instead raises ValueError on any error
        msg_printer = messages.MessageExceptionRaiser()
        msg_handler = messages.MessageHandler(msg_printer)

        input_stream = InputStream(expression)

        parsed_tree = sa_systemrdl.parse(
            input_stream,
            "eval_expr_root",
            messages.RdlSaErrorListener(msg_handler)
        )

        visitor = ExprVisitor(self)

        # override visitor to use local message handler
        visitor.msg = msg_handler

        result = visitor.visit(parsed_tree)
        result.predict_type()
        return result.get_value()


class RDLEnvironment:
    """
    Container object for misc resources that are preserved outside the lifetime
    of source compilation
    """
    def __init__(self, args_dict: Dict[str, Any]):

        # Collect args
        message_printer = args_dict.pop('message_printer', messages.MessagePrinter())
        w_flags = args_dict.pop('warning_flags', 0)
        e_flags = args_dict.pop('error_flags', 0)
        self.dedent_desc = args_dict.pop('dedent_desc', True)
        self.use_extended_type_name_gen = args_dict.pop('extended_dpa_type_names', True)
        self.perl_safe_opcodes = args_dict.pop('perl_safe_opcodes', [
            ':base_core', ':base_mem', ':base_loop', ':base_orig', ':base_math',
            ':base_thread', ':filesys_read', ':sys_db', ':load',
            'sort', 'tied', 'pack', 'unpack', 'reset'
        ])

        self.chk_missing_reset = self.chk_flag_severity(warnings.MISSING_RESET, w_flags, e_flags)
        self.chk_implicit_field_pos = self.chk_flag_severity(warnings.IMPLICIT_FIELD_POS, w_flags, e_flags)
        self.chk_implicit_addr = self.chk_flag_severity(warnings.IMPLICIT_ADDR, w_flags, e_flags)
        self.chk_stride_not_pow2 = self.chk_flag_severity(warnings.STRIDE_NOT_POW2, w_flags, e_flags)
        self.chk_strict_self_align = self.chk_flag_severity(warnings.STRICT_SELF_ALIGN, w_flags, e_flags)
        self.chk_sparse_reg_stride = self.chk_flag_severity(warnings.SPARSE_REG_STRIDE, w_flags, e_flags)

        self.msg = messages.MessageHandler(message_printer)
        self.property_rules = PropertyRuleBook(self)

    @staticmethod
    def chk_flag_severity(flag: int, w_flags: int, e_flags: int) -> messages.Severity:
        if bool(e_flags & flag):
            return messages.Severity.ERROR
        elif bool(w_flags & flag):
            return messages.Severity.WARNING
        else:
            return messages.Severity.NONE



from collections import OrderedDict
from typing import TYPE_CHECKING

from ..parser.SystemRDLParser import SystemRDLParser

from .BaseVisitor import BaseVisitor
from .helpers import get_ID_text

from .. import component as comp
from ..source_ref import src_ref_from_antlr
from .. import rdltypes

if TYPE_CHECKING:
    from ..compiler import RDLCompiler

class StructVisitor(BaseVisitor):

    def __init__(self, compiler: 'RDLCompiler', parent_component: 'comp.Component') -> None:
        super().__init__(compiler)
        self.parent_component = parent_component

    def visitStruct_def(self, ctx: SystemRDLParser.Struct_defContext):
        self.compiler.namespace.enter_scope()

        is_abstract = ctx.ABSTRACT_kw() is not None
        struct_name = get_ID_text(ctx.name)

        if ctx.base is not None:
            # Get the base struct type
            base_name = get_ID_text(ctx.base)
            base_type = self.compiler.namespace.lookup_type(base_name)
            if base_type is None:
                self.msg.fatal(
                    "Type '%s' is not defined" % base_name,
                    src_ref_from_antlr(ctx.base)
                )
            if not rdltypes.is_user_struct(base_type):
                self.msg.fatal(
                    "Base type '%s' is not a struct" % base_name,
                    src_ref_from_antlr(ctx.base)
                )
        else:
            base_type = rdltypes.UserStruct


        # Collect struct member elements
        members = OrderedDict()
        for struct_elem_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Struct_elemContext):
            member_type, member_name, member_src_ref = self.visit(struct_elem_ctx)

            if member_name in members:
                self.msg.error(
                    "Struct member '%s' has already been defined" % member_name,
                    member_src_ref
                )
                continue

            if member_name in base_type._members:
                self.msg.error(
                    "Struct member '%s' has already been defined in base struct '%s'"
                    % (member_name, base_type.__name__),
                    member_src_ref
                )
                continue

            members[member_name] = member_type


        # Create Struct type
        struct_type = base_type.define_new(struct_name, members, is_abstract, self.parent_component)

        self.compiler.namespace.exit_scope()
        return struct_type, struct_name, src_ref_from_antlr(ctx.name)


    def visitStruct_elem(self, ctx: SystemRDLParser.Struct_elemContext):

        member_name = get_ID_text(ctx.ID())
        member_src_ref = src_ref_from_antlr(ctx.ID())

        member_type = self.visit(ctx.struct_type())

        if ctx.array_type_suffix() is not None:
            member_type = rdltypes.ArrayedType(member_type)

        return member_type, member_name, member_src_ref


    _CompType_Map = {
        SystemRDLParser.FIELD_kw    : comp.Field,
        SystemRDLParser.REG_kw      : comp.Reg,
        SystemRDLParser.REGFILE_kw  : comp.Regfile,
        SystemRDLParser.ADDRMAP_kw  : comp.Addrmap,
        SystemRDLParser.SIGNAL_kw   : comp.Signal,
        SystemRDLParser.MEM_kw      : comp.Mem
    }
    def visitStruct_type(self, ctx: SystemRDLParser.Struct_typeContext):
        if ctx.data_type() is not None:
            data_type_token = self.visit(ctx.data_type())
            member_type = self.datatype_from_token(data_type_token)
        elif ctx.component_type() is not None:
            type_token = self.visit(ctx.component_type())
            member_type = self._CompType_Map[type_token.type]
        else:
            raise RuntimeError

        return member_type



from typing import Optional, Any, List, TYPE_CHECKING
import re

from .compiler import RDLCompiler
from .source_ref import FileSourceRef, SourceRefBase
from . import component as comp

if TYPE_CHECKING:
    from typing import TypeVar, Type
    ComponentClass = TypeVar('ComponentClass', bound=comp.Component)
    AddressableComponentClass = TypeVar('AddressableComponentClass', bound=comp.AddressableComponent)

class RDLImporter:
    def __init__(self, compiler: RDLCompiler):
        """
        Base class for external parsers to import data into the register model

        Parameters
        ----------
        compiler: :class:`RDLCompiler`
            Reference to compiler instance where the import is being performed.
        """

        #: Reference to the current compiler instance
        self.compiler = compiler

        #: Reference to the compiler message handler
        #:
        #: .. warning::
        #:
        #:      This will be deprecated in a future release. See this page for more details: https://github.com/SystemRDL/systemrdl-compiler/issues/168
        self.msg = compiler.env.msg

        #: Source reference used by default when not specified in importer
        #: model construction methods
        #: By default, this simply points to the file currently being imported,
        #: without any line offset information
        self.default_src_ref = None # type: SourceRefBase

    def import_file(self, path: str) -> None:
        """
        Importer entry point.

        Extend this function to read the file and perform the import.
        Be sure to call ``super().import_file(path)``

        Parameters
        ----------
        path: str
            Path to file
        """
        self.default_src_ref = FileSourceRef(path)

    # ------------------------ create_xxx_definition() -------------------------
    def _create_definition(self, cls: "Type[ComponentClass]", type_name: Optional[str] = None, src_ref: Optional[SourceRefBase] = None) -> "ComponentClass":
        if (type_name is not None) and not re.fullmatch(r"[a-zA-Z_]\w*", type_name):
            raise ValueError(f"Type name has invalid characters: '{type_name}'")
        C = cls()
        C.type_name = type_name
        C.def_src_ref = src_ref or self.default_src_ref
        return C

    def create_field_definition(self, type_name: Optional[str] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Field:
        """
        Parameters
        ----------
        type_name: str
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Field`
            Component definition
        """
        return self._create_definition(comp.Field, type_name, src_ref)

    def create_reg_definition(self, type_name: Optional[str] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Reg:
        """
        Parameters
        ----------
        type_name: str
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Reg`
            Component definition
        """
        return self._create_definition(comp.Reg, type_name, src_ref)

    def create_regfile_definition(self, type_name: Optional[str] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Regfile:
        """
        Parameters
        ----------
        type_name: str
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Regfile`
            Component definition
        """
        return self._create_definition(comp.Regfile, type_name, src_ref)

    def create_addrmap_definition(self, type_name: Optional[str] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Addrmap:
        """
        Parameters
        ----------
        type_name: str
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Addrmap`
            Component definition
        """
        return self._create_definition(comp.Addrmap, type_name, src_ref)

    def create_mem_definition(self, type_name: Optional[str] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Mem:
        """
        Parameters
        ----------
        type_name: str
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Mem`
            Component definition
        """
        C = self._create_definition(comp.Mem, type_name, src_ref)
        C.external = True
        return C


    # --------------------------- instantiate_xxx() ----------------------------
    def _instantiate(self, comp_def: "ComponentClass", inst_name: str, src_ref: Optional[SourceRefBase] = None) -> "ComponentClass":
        if comp_def.is_instance:
            raise ValueError("Component is already instantiated")
        if not re.fullmatch(r"[a-zA-Z_]\w*", inst_name):
            raise ValueError(f"Instance name has invalid characters: '{inst_name}'")

        if comp_def.type_name is None:
            # anonymous definition. No need to copy since it can only be used once
            comp_inst = comp_def
        else:
            # named definition. Copy so that the definition can be re-used
            comp_inst = comp_def._copy_for_inst({})
            comp_inst.original_def = comp_def

        comp_inst.is_instance = True
        comp_inst.inst_name = inst_name
        comp_inst.inst_src_ref = src_ref or self.default_src_ref
        return comp_inst

    def instantiate_field(self, comp_def: comp.Field, inst_name: str, bit_offset: int, bit_width: int, src_ref: Optional[SourceRefBase] = None) -> comp.Field:
        """
        Parameters
        ----------
        comp_def: :class:`comp.Field`
        inst_name: str
        bit_offset: int
        bit_width: int
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Field`
            Component instance
        """
        assert isinstance(comp_def, comp.Field)
        comp_inst = self._instantiate(comp_def, inst_name, src_ref)
        comp_inst.low = bit_offset
        comp_inst.high = bit_offset + bit_width - 1
        comp_inst.lsb = comp_inst.low
        comp_inst.msb = comp_inst.high
        comp_inst.width = bit_width
        return comp_inst

    def _instantiate_addressable(self, comp_def: "AddressableComponentClass", inst_name: str, addr_offset: int, array_dimensions: Optional[List[int]] = None, array_stride: Optional[int] = None, src_ref: Optional[SourceRefBase] = None) -> "AddressableComponentClass":

        comp_inst = self._instantiate(comp_def, inst_name, src_ref)
        comp_inst.addr_offset = addr_offset
        if array_dimensions:
            comp_inst.is_array = True
            comp_inst.array_dimensions = array_dimensions
            comp_inst.array_stride = array_stride
        return comp_inst

    def instantiate_reg(self, comp_def: comp.Reg, inst_name: str, addr_offset: int, array_dimensions: Optional[List[int]] = None, array_stride: Optional[int] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Reg:
        """
        Parameters
        ----------
        comp_def: :class:`comp.Reg`
        inst_name: str
        addr_offset: int
        array_dimensions: int
        array_stride: int
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Field`
            Component instance
        """
        assert isinstance(comp_def, comp.Reg)
        return self._instantiate_addressable(comp_def, inst_name, addr_offset, array_dimensions, array_stride, src_ref)

    def instantiate_regfile(self, comp_def: comp.Regfile, inst_name: str, addr_offset: int, array_dimensions: Optional[List[int]] = None, array_stride: Optional[int] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Regfile:
        """
        Parameters
        ----------
        comp_def: :class:`comp.Regfile`
        inst_name: str
        addr_offset: int
        array_dimensions: int
        array_stride: int
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Field`
            Component instance
        """
        assert isinstance(comp_def, comp.Regfile)
        return self._instantiate_addressable(comp_def, inst_name, addr_offset, array_dimensions, array_stride, src_ref)

    def instantiate_addrmap(self, comp_def: comp.Addrmap, inst_name: str, addr_offset: int, array_dimensions: Optional[List[int]] = None, array_stride: Optional[int] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Addrmap:
        """
        Parameters
        ----------
        comp_def: :class:`comp.Addrmap`
        inst_name: str
        addr_offset: int
        array_dimensions: int
        array_stride: int
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Field`
            Component instance
        """
        assert isinstance(comp_def, comp.Addrmap)
        return self._instantiate_addressable(comp_def, inst_name, addr_offset, array_dimensions, array_stride, src_ref)

    def instantiate_mem(self, comp_def: comp.Mem, inst_name: str, addr_offset: int, array_dimensions: Optional[List[int]] = None, array_stride: Optional[int] = None, src_ref: Optional[SourceRefBase] = None) -> comp.Mem:
        """
        Parameters
        ----------
        comp_def: :class:`comp.Mem`
        inst_name: str
        addr_offset: int
        array_dimensions: int
        array_stride: int
        src_ref: :class:`~SourceRefBase`

        Returns
        -------
        :class:`~comp.Field`
            Component instance
        """
        assert isinstance(comp_def, comp.Mem)
        return self._instantiate_addressable(comp_def, inst_name, addr_offset, array_dimensions, array_stride, src_ref)

    #---------------------------------------------------------------------------
    def add_child(self, parent: comp.Component, child: comp.Component) -> None:
        """
        Add a child component instance to an existing parent definition or instance

        .. versionadded:: 1.16
        """
        bad = True
        if isinstance(parent, comp.Addrmap):
            if isinstance(child, comp.AddressableComponent):
                bad = False
        elif isinstance(parent, comp.Regfile):
            if isinstance(child, (comp.Regfile, comp.Reg)):
                bad = False
        elif isinstance(parent, comp.Mem):
            if isinstance(child, comp.Reg):
                bad = False
        elif isinstance(parent, comp.Reg):
            if isinstance(child, comp.Field):
                bad = False
        if bad:
            raise TypeError(f"Parent {repr(parent)} cannot be assigned child {repr(child)}")

        if not child.is_instance:
            raise ValueError("Child must be an instance if adding to a parent")

        parent.children.append(child)

    def assign_property(self, component: comp.Component, prop_name: str, value: Any, src_ref: Optional[SourceRefBase] = None) -> None:
        """
        Assign a property to a component.

        Parameters
        ----------
        prop_name: str
            Name of the SystemRDL property
        value:
            Value to assign
        src_ref: :class:`~SourceRefBase`
            Optionally provide a more detailed source reference object associated
            with this assignment
        """

        rule = self.compiler.env.property_rules.lookup_property(prop_name)
        if rule is None:
            raise ValueError(f"Unrecognized property '{prop_name}'")
        if isinstance(value, comp.Component):
            # Warn user that this is not how references work
            raise TypeError("Invalid assignment type")
        src_ref = src_ref or self.default_src_ref
        rule.assign_value(component, value, src_ref)

    def register_root_component(self, definition: comp.Component) -> None:
        """
        Register a component definition with the root namespace so that it is
        visible to other compiled files.

        Parameters
        ----------
        definition: :class:`comp.Component`
            Component definition to register

        """
        if definition.type_name is None:
            raise ValueError("Component must have a type_name")

        if definition.is_instance:
            raise ValueError("Component cannot already be instantiated")

        if len(self.compiler.namespace.type_ns_stack) != 1:
            raise RuntimeError("Namespace was not properly reset to root scope")

        # Register with namespace
        self.compiler.namespace.register_type(
            definition.type_name,
            definition,
            definition.def_src_ref
        )

        # Add to root component definition list
        self.compiler.root.comp_defs[definition.type_name] = definition

    def lookup_root_component(self, type_name: str) -> Optional[comp.Component]:
        """
        Lookup a component definition from the root namespace. If the type name
        is not defined, returns ``None``
        """
        return self.compiler.root.comp_defs.get(type_name, None)



from antlr4 import *
from _typeshed import Incomplete
from io import StringIO as StringIO
from typing import TextIO

def serializedATN(): ...

class SystemRDLParser(Parser):
    grammarFileName: str
    atn: Incomplete
    decisionsToDFA: Incomplete
    sharedContextCache: Incomplete
    literalNames: Incomplete
    symbolicNames: Incomplete
    RULE_root: int
    RULE_eval_expr_root: int
    RULE_root_elem: int
    RULE_component_def: int
    RULE_explicit_component_inst: int
    RULE_component_inst_alias: int
    RULE_component_named_def: int
    RULE_component_anon_def: int
    RULE_component_body: int
    RULE_component_body_elem: int
    RULE_component_insts: int
    RULE_component_inst: int
    RULE_field_inst_reset: int
    RULE_inst_addr_fixed: int
    RULE_inst_addr_stride: int
    RULE_inst_addr_align: int
    RULE_component_inst_type: int
    RULE_component_type: int
    RULE_component_type_primary: int
    RULE_param_def: int
    RULE_param_def_elem: int
    RULE_param_inst: int
    RULE_param_assignment: int
    RULE_expr: int
    RULE_expr_primary: int
    RULE_concatenate: int
    RULE_replicate: int
    RULE_paren_expr: int
    RULE_cast: int
    RULE_cast_width_expr: int
    RULE_range_suffix: int
    RULE_array_suffix: int
    RULE_array_type_suffix: int
    RULE_data_type: int
    RULE_basic_data_type: int
    RULE_literal: int
    RULE_number: int
    RULE_string_literal: int
    RULE_boolean_literal: int
    RULE_array_literal: int
    RULE_struct_literal: int
    RULE_struct_kv: int
    RULE_enum_literal: int
    RULE_accesstype_literal: int
    RULE_onreadtype_literal: int
    RULE_onwritetype_literal: int
    RULE_addressingtype_literal: int
    RULE_precedencetype_literal: int
    RULE_instance_ref: int
    RULE_instance_ref_element: int
    RULE_prop_ref: int
    RULE_local_property_assignment: int
    RULE_dynamic_property_assignment: int
    RULE_normal_prop_assign: int
    RULE_encode_prop_assign: int
    RULE_prop_mod_assign: int
    RULE_prop_assignment_rhs: int
    RULE_prop_keyword: int
    RULE_prop_mod: int
    RULE_udp_def: int
    RULE_udp_attr: int
    RULE_udp_type: int
    RULE_udp_data_type: int
    RULE_udp_usage: int
    RULE_udp_comp_type: int
    RULE_udp_default: int
    RULE_udp_constraint: int
    RULE_enum_def: int
    RULE_enum_entry: int
    RULE_enum_prop_assign: int
    RULE_struct_def: int
    RULE_struct_elem: int
    RULE_struct_type: int
    RULE_constraint_def: int
    RULE_constraint_named_def: int
    RULE_constraint_anon_def: int
    RULE_constraint_body: int
    RULE_constraint_body_elem: int
    RULE_constraint_insts: int
    RULE_constr_relational: int
    RULE_constr_prop_assign: int
    RULE_constr_inside_values: int
    RULE_constr_inside_enum: int
    RULE_constr_lhs: int
    RULE_constr_inside_value: int
    ruleNames: Incomplete
    EOF: Incomplete
    T__0: int
    T__1: int
    T__2: int
    T__3: int
    T__4: int
    T__5: int
    T__6: int
    T__7: int
    T__8: int
    T__9: int
    T__10: int
    T__11: int
    T__12: int
    T__13: int
    T__14: int
    SL_COMMENT: int
    ML_COMMENT: int
    BOOLEAN_kw: int
    BIT_kw: int
    LONGINT_kw: int
    UNSIGNED_kw: int
    STRING_kw: int
    ACCESSTYPE_kw: int
    ADDRESSINGTYPE_kw: int
    ONREADTYPE_kw: int
    ONWRITETYPE_kw: int
    ALIAS_kw: int
    EXTERNAL_kw: int
    INTERNAL_kw: int
    ADDRMAP_kw: int
    REGFILE_kw: int
    REG_kw: int
    FIELD_kw: int
    MEM_kw: int
    SIGNAL_kw: int
    TRUE_kw: int
    FALSE_kw: int
    NA_kw: int
    RW_kw: int
    WR_kw: int
    R_kw: int
    W_kw: int
    RW1_kw: int
    W1_kw: int
    RCLR_kw: int
    RSET_kw: int
    RUSER_kw: int
    WOSET_kw: int
    WOCLR_kw: int
    WOT_kw: int
    WZS_kw: int
    WZC_kw: int
    WZT_kw: int
    WCLR_kw: int
    WSET_kw: int
    WUSER_kw: int
    COMPACT_kw: int
    REGALIGN_kw: int
    FULLALIGN_kw: int
    HW_kw: int
    SW_kw: int
    POSEDGE_kw: int
    NEGEDGE_kw: int
    BOTHEDGE_kw: int
    LEVEL_kw: int
    NONSTICKY_kw: int
    ABSTRACT_kw: int
    ALL_kw: int
    COMPONENT_kw: int
    COMPONENTWIDTH_kw: int
    CONSTRAINT_kw: int
    DEFAULT_kw: int
    ENUM_kw: int
    ENCODE_kw: int
    INSIDE_kw: int
    NUMBER_kw: int
    PROPERTY_kw: int
    REF_kw: int
    STRUCT_kw: int
    THIS_kw: int
    TYPE_kw: int
    ALTERNATE_kw: int
    BYTE_kw: int
    INT_kw: int
    PRECEDENCETYPE_kw: int
    REAL_kw: int
    SHORTINT_kw: int
    SHORTREAL_kw: int
    SIGNED_kw: int
    WITH_kw: int
    WITHIN_kw: int
    INT: int
    HEX_INT: int
    VLOG_INT: int
    STRING: int
    PLUS: int
    MINUS: int
    BNOT: int
    NOT: int
    BAND: int
    NAND: int
    AND: int
    OR: int
    BOR: int
    NOR: int
    XOR: int
    XNOR: int
    LSHIFT: int
    RSHIFT: int
    MULT: int
    EXP: int
    DIV: int
    MOD: int
    EQ: int
    ASSIGN: int
    NEQ: int
    LEQ: int
    LT: int
    GEQ: int
    GT: int
    AT: int
    INC: int
    ALIGN: int
    WS: int
    ID: int
    def __init__(self, input: TokenStream, output: TextIO = ...) -> None: ...
    class RootContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def EOF(self): ...
        def root_elem(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    state: int
    def root(self): ...
    class Eval_expr_rootContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def EOF(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def eval_expr_root(self): ...
    class Root_elemContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_def(self): ...
        def enum_def(self): ...
        def udp_def(self): ...
        def struct_def(self): ...
        def constraint_def(self): ...
        def explicit_component_inst(self): ...
        def local_property_assignment(self): ...
        def dynamic_property_assignment(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def root_elem(self): ...
    class Component_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_named_def(self): ...
        def component_inst_type(self): ...
        def component_insts(self): ...
        def component_anon_def(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_def(self): ...
    class Explicit_component_instContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def component_insts(self): ...
        def component_inst_type(self): ...
        def component_inst_alias(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def explicit_component_inst(self): ...
    class Component_inst_aliasContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ALIAS_kw(self): ...
        def ID(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_inst_alias(self): ...
    class Component_named_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_type(self): ...
        def ID(self): ...
        def component_body(self): ...
        def param_def(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_named_def(self): ...
    class Component_anon_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_type(self): ...
        def component_body(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_anon_def(self): ...
    class Component_bodyContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_body_elem(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_body(self): ...
    class Component_body_elemContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_def(self): ...
        def enum_def(self): ...
        def struct_def(self): ...
        def constraint_def(self): ...
        def explicit_component_inst(self): ...
        def local_property_assignment(self): ...
        def dynamic_property_assignment(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_body_elem(self): ...
    class Component_instsContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_inst(self, i: int = None): ...
        def param_inst(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_insts(self): ...
    class Component_instContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def range_suffix(self): ...
        def field_inst_reset(self): ...
        def inst_addr_fixed(self): ...
        def inst_addr_stride(self): ...
        def inst_addr_align(self): ...
        def array_suffix(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_inst(self): ...
    class Field_inst_resetContext(ParserRuleContext):
        parser: Incomplete
        op: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def ASSIGN(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def field_inst_reset(self): ...
    class Inst_addr_fixedContext(ParserRuleContext):
        parser: Incomplete
        op: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def AT(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def inst_addr_fixed(self): ...
    class Inst_addr_strideContext(ParserRuleContext):
        parser: Incomplete
        op: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def INC(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def inst_addr_stride(self): ...
    class Inst_addr_alignContext(ParserRuleContext):
        parser: Incomplete
        op: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def ALIGN(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def inst_addr_align(self): ...
    class Component_inst_typeContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def EXTERNAL_kw(self): ...
        def INTERNAL_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_inst_type(self): ...
    class Component_typeContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_type_primary(self): ...
        def SIGNAL_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_type(self): ...
    class Component_type_primaryContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ADDRMAP_kw(self): ...
        def REGFILE_kw(self): ...
        def REG_kw(self): ...
        def FIELD_kw(self): ...
        def MEM_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def component_type_primary(self): ...
    class Param_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def param_def_elem(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def param_def(self): ...
    class Param_def_elemContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def data_type(self): ...
        def ID(self): ...
        def array_type_suffix(self): ...
        def ASSIGN(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def param_def_elem(self): ...
    class Param_instContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def param_assignment(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def param_inst(self): ...
    class Param_assignmentContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def param_assignment(self): ...
    class ExprContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def getRuleIndex(self): ...
        def copyFrom(self, ctx: ParserRuleContext): ...
    class BinaryExprContext(ExprContext):
        op: Incomplete
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def expr(self, i: int = None): ...
        def EXP(self): ...
        def MULT(self): ...
        def DIV(self): ...
        def MOD(self): ...
        def PLUS(self): ...
        def MINUS(self): ...
        def LSHIFT(self): ...
        def RSHIFT(self): ...
        def LT(self): ...
        def LEQ(self): ...
        def GT(self): ...
        def GEQ(self): ...
        def EQ(self): ...
        def NEQ(self): ...
        def AND(self): ...
        def XOR(self): ...
        def XNOR(self): ...
        def OR(self): ...
        def BAND(self): ...
        def BOR(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    class UnaryExprContext(ExprContext):
        op: Incomplete
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def expr_primary(self): ...
        def PLUS(self): ...
        def MINUS(self): ...
        def BNOT(self): ...
        def NOT(self): ...
        def AND(self): ...
        def NAND(self): ...
        def OR(self): ...
        def NOR(self): ...
        def XOR(self): ...
        def XNOR(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    class NOPContext(ExprContext):
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def expr_primary(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    class TernaryExprContext(ExprContext):
        op: Incomplete
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def expr(self, i: int = None): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def expr(self, _p: int = 0): ...
    class Expr_primaryContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def literal(self): ...
        def concatenate(self): ...
        def replicate(self): ...
        def paren_expr(self): ...
        def cast(self): ...
        def prop_ref(self): ...
        def instance_ref(self): ...
        def struct_literal(self): ...
        def array_literal(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def expr_primary(self): ...
    class ConcatenateContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def concatenate(self): ...
    class ReplicateContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def concatenate(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def replicate(self): ...
    class Paren_exprContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def paren_expr(self): ...
    class CastContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def getRuleIndex(self): ...
        def copyFrom(self, ctx: ParserRuleContext): ...
    class CastWidthContext(CastContext):
        op: Incomplete
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def cast_width_expr(self): ...
        def expr(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    class CastTypeContext(CastContext):
        typ: Incomplete
        op: Incomplete
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def expr(self): ...
        def BOOLEAN_kw(self): ...
        def BIT_kw(self): ...
        def LONGINT_kw(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def cast(self): ...
    class Cast_width_exprContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def literal(self): ...
        def paren_expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def cast_width_expr(self): ...
    class Range_suffixContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def range_suffix(self): ...
    class Array_suffixContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def array_suffix(self): ...
    class Array_type_suffixContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def array_type_suffix(self): ...
    class Data_typeContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def basic_data_type(self): ...
        def ACCESSTYPE_kw(self): ...
        def ADDRESSINGTYPE_kw(self): ...
        def ONREADTYPE_kw(self): ...
        def ONWRITETYPE_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def data_type(self): ...
    class Basic_data_typeContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def BIT_kw(self): ...
        def LONGINT_kw(self): ...
        def UNSIGNED_kw(self): ...
        def STRING_kw(self): ...
        def BOOLEAN_kw(self): ...
        def ID(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def basic_data_type(self): ...
    class LiteralContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def number(self): ...
        def string_literal(self): ...
        def boolean_literal(self): ...
        def accesstype_literal(self): ...
        def onreadtype_literal(self): ...
        def onwritetype_literal(self): ...
        def addressingtype_literal(self): ...
        def precedencetype_literal(self): ...
        def enum_literal(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def literal(self): ...
    class NumberContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def getRuleIndex(self): ...
        def copyFrom(self, ctx: ParserRuleContext): ...
    class NumberHexContext(NumberContext):
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def HEX_INT(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    class NumberVerilogContext(NumberContext):
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def VLOG_INT(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    class NumberIntContext(NumberContext):
        def __init__(self, parser, ctx: ParserRuleContext) -> None: ...
        def INT(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def number(self): ...
    class String_literalContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def STRING(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def string_literal(self): ...
    class Boolean_literalContext(ParserRuleContext):
        parser: Incomplete
        val: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def TRUE_kw(self): ...
        def FALSE_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def boolean_literal(self): ...
    class Array_literalContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def array_literal(self): ...
    class Struct_literalContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def struct_kv(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def struct_literal(self): ...
    class Struct_kvContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def struct_kv(self): ...
    class Enum_literalContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def enum_literal(self): ...
    class Accesstype_literalContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def NA_kw(self): ...
        def RW_kw(self): ...
        def WR_kw(self): ...
        def R_kw(self): ...
        def W_kw(self): ...
        def RW1_kw(self): ...
        def W1_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def accesstype_literal(self): ...
    class Onreadtype_literalContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def RCLR_kw(self): ...
        def RSET_kw(self): ...
        def RUSER_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def onreadtype_literal(self): ...
    class Onwritetype_literalContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def WOSET_kw(self): ...
        def WOCLR_kw(self): ...
        def WOT_kw(self): ...
        def WZS_kw(self): ...
        def WZC_kw(self): ...
        def WZT_kw(self): ...
        def WCLR_kw(self): ...
        def WSET_kw(self): ...
        def WUSER_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def onwritetype_literal(self): ...
    class Addressingtype_literalContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def COMPACT_kw(self): ...
        def REGALIGN_kw(self): ...
        def FULLALIGN_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def addressingtype_literal(self): ...
    class Precedencetype_literalContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def HW_kw(self): ...
        def SW_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def precedencetype_literal(self): ...
    class Instance_refContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def instance_ref_element(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def instance_ref(self): ...
    class Instance_ref_elementContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def array_suffix(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def instance_ref_element(self): ...
    class Prop_refContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def instance_ref(self): ...
        def prop_keyword(self): ...
        def ID(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def prop_ref(self): ...
    class Local_property_assignmentContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def normal_prop_assign(self): ...
        def DEFAULT_kw(self): ...
        def encode_prop_assign(self): ...
        def prop_mod_assign(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def local_property_assignment(self): ...
    class Dynamic_property_assignmentContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def instance_ref(self): ...
        def normal_prop_assign(self): ...
        def encode_prop_assign(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def dynamic_property_assignment(self): ...
    class Normal_prop_assignContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def prop_keyword(self): ...
        def ID(self): ...
        def ASSIGN(self): ...
        def prop_assignment_rhs(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def normal_prop_assign(self): ...
    class Encode_prop_assignContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ENCODE_kw(self): ...
        def ASSIGN(self): ...
        def ID(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def encode_prop_assign(self): ...
    class Prop_mod_assignContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def prop_mod(self): ...
        def ID(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def prop_mod_assign(self): ...
    class Prop_assignment_rhsContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def precedencetype_literal(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def prop_assignment_rhs(self): ...
    class Prop_keywordContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def SW_kw(self): ...
        def HW_kw(self): ...
        def RCLR_kw(self): ...
        def RSET_kw(self): ...
        def WOCLR_kw(self): ...
        def WOSET_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def prop_keyword(self): ...
    class Prop_modContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def POSEDGE_kw(self): ...
        def NEGEDGE_kw(self): ...
        def BOTHEDGE_kw(self): ...
        def LEVEL_kw(self): ...
        def NONSTICKY_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def prop_mod(self): ...
    class Udp_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def PROPERTY_kw(self): ...
        def ID(self): ...
        def udp_attr(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_def(self): ...
    class Udp_attrContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def udp_type(self): ...
        def udp_usage(self): ...
        def udp_default(self): ...
        def udp_constraint(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_attr(self): ...
    class Udp_typeContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def TYPE_kw(self): ...
        def ASSIGN(self): ...
        def udp_data_type(self): ...
        def array_type_suffix(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_type(self): ...
    class Udp_data_typeContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_type_primary(self): ...
        def REF_kw(self): ...
        def NUMBER_kw(self): ...
        def basic_data_type(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_data_type(self): ...
    class Udp_usageContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def COMPONENT_kw(self): ...
        def ASSIGN(self): ...
        def udp_comp_type(self, i: int = None): ...
        def OR(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_usage(self): ...
    class Udp_comp_typeContext(ParserRuleContext):
        parser: Incomplete
        kw: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def component_type(self): ...
        def CONSTRAINT_kw(self): ...
        def ALL_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_comp_type(self): ...
    class Udp_defaultContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def DEFAULT_kw(self): ...
        def ASSIGN(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_default(self): ...
    class Udp_constraintContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def CONSTRAINT_kw(self): ...
        def ASSIGN(self): ...
        def COMPONENTWIDTH_kw(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def udp_constraint(self): ...
    class Enum_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ENUM_kw(self): ...
        def ID(self): ...
        def enum_entry(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def enum_def(self): ...
    class Enum_entryContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def ASSIGN(self): ...
        def expr(self): ...
        def enum_prop_assign(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def enum_entry(self): ...
    class Enum_prop_assignContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def ASSIGN(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def enum_prop_assign(self): ...
    class Struct_defContext(ParserRuleContext):
        parser: Incomplete
        name: Incomplete
        base: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def STRUCT_kw(self): ...
        def ID(self, i: int = None): ...
        def ABSTRACT_kw(self): ...
        def struct_elem(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def struct_def(self): ...
    class Struct_elemContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def struct_type(self): ...
        def ID(self): ...
        def array_type_suffix(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def struct_elem(self): ...
    class Struct_typeContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def data_type(self): ...
        def component_type(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def struct_type(self): ...
    class Constraint_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def constraint_named_def(self): ...
        def constraint_insts(self): ...
        def constraint_anon_def(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constraint_def(self): ...
    class Constraint_named_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def CONSTRAINT_kw(self): ...
        def ID(self): ...
        def constraint_body(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constraint_named_def(self): ...
    class Constraint_anon_defContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def CONSTRAINT_kw(self): ...
        def constraint_body(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constraint_anon_def(self): ...
    class Constraint_bodyContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def constraint_body_elem(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constraint_body(self): ...
    class Constraint_body_elemContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def constr_relational(self): ...
        def constr_prop_assign(self): ...
        def constr_inside_values(self): ...
        def constr_inside_enum(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constraint_body_elem(self): ...
    class Constraint_instsContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constraint_insts(self): ...
    class Constr_relationalContext(ParserRuleContext):
        parser: Incomplete
        op: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self, i: int = None): ...
        def LT(self): ...
        def LEQ(self): ...
        def GT(self): ...
        def GEQ(self): ...
        def EQ(self): ...
        def NEQ(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constr_relational(self): ...
    class Constr_prop_assignContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def ID(self): ...
        def ASSIGN(self): ...
        def expr(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constr_prop_assign(self): ...
    class Constr_inside_valuesContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def constr_lhs(self): ...
        def INSIDE_kw(self): ...
        def constr_inside_value(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constr_inside_values(self): ...
    class Constr_inside_enumContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def constr_lhs(self): ...
        def INSIDE_kw(self): ...
        def ID(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constr_inside_enum(self): ...
    class Constr_lhsContext(ParserRuleContext):
        parser: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def THIS_kw(self): ...
        def instance_ref(self): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constr_lhs(self): ...
    class Constr_inside_valueContext(ParserRuleContext):
        parser: Incomplete
        val: Incomplete
        l_val: Incomplete
        r_val: Incomplete
        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1) -> None: ...
        def expr(self, i: int = None): ...
        def getRuleIndex(self): ...
        def accept(self, visitor: ParseTreeVisitor): ...
    def constr_inside_value(self): ...
    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int): ...
    def expr_sempred(self, localctx: ExprContext, predIndex: int): ...



import re
from typing import List, Union, Tuple, TYPE_CHECKING, Optional, Match, Dict

from ..source_ref import SourceRefBase, SegmentedSourceRef
from . import segment_map

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment

class VerilogPreprocessor:
    def __init__(self, env: 'RDLEnvironment', text: str, src_seg_map: Optional[segment_map.SegmentMap]=None, src_ref_override: Optional[SourceRefBase]=None, defines: Optional[Dict[str,str]]=None):
        self.env = env

        # Unprocessed text
        self._text = text

        # Top-level preprocessing pass requires the segment map in order to
        # generate source references.
        # Once the preprocessor recurses during macro expansion, this is not
        # used, and instead the src_ref_override is passed along statically
        self._src_seg_map = src_seg_map

        # Preprocessor macro expansions result in a few recursive calls to the
        # verilog preprocessor. After the first recursion, it is not practical
        # to keep track of character offsets anymore, so instead a blanket
        # src_ref is used for any errors that may occur while recursing.
        # This is the alternate Source Ref to use for this level of recursion
        self._src_ref_override = src_ref_override

        # Current character offset within text
        self._scan_idx = 0

        # Start offset of current segment selection
        self._seg_start_idx = 0

        # Macro namespace
        self._macro_defs = {} # type: Dict[str, Macro]
        if defines is not None:
            for k, v in defines.items():
                macro = Macro(v, [])
                self._macro_defs[k] = macro

        self._conditional = ConditionalState()
        self._conditional_stack = [] # type: List[ConditionalState]

        # Stack of what macros are currently being processed
        self._active_macro_stack = [] # type: List[str]

        self._output_text_segments = [] # type: List[str]
        self._current_output_idx = 0
        self._output_seg_map = None # type: Optional[segment_map.SegmentMap]
        if self._src_seg_map:
            self._output_seg_map = segment_map.SegmentMap()

    def preprocess(self) -> Tuple[str, Optional[segment_map.SegmentMap]]:
        self.main_scanner()

        if not self._conditional.is_idle or self._conditional_stack:
            self.env.msg.fatal(
                "Incomplete conditional. Missing `endif",
                self.get_err_src_ref(len(self._text), len(self._text))
            )

        output_text = "".join(self._output_text_segments)
        return (output_text, self._output_seg_map)

    def get_err_src_ref(self, start:int, end:int) -> SourceRefBase:
        """
        When emitting an error, return the appropriate src_ref.
        If at the top-level preprocessing pass, creates a unique reference
        against src_seg_map. Otherwise returns src_ref_override
        """
        if self._src_ref_override:
            return self._src_ref_override
        assert isinstance(self._src_seg_map, segment_map.SegmentMap)
        return SegmentedSourceRef(self._src_seg_map, start, end)

    #---------------------------------------------------------------------------
    # Main text scanner
    #---------------------------------------------------------------------------
    def main_scanner(self) -> None:

        queries = [
            # Skip comments
            ('mlc', r'/\*.*?\*/'),
            ('slc', r'//.*?$'),

            # Skip strings
            ('str', r'"(\\"|\\\\|[^\\"])*"'),

            # Conditionals
            ('ifdef', r'^[ \t]*(`ifdef)[ \t]+(\w+)\b'),
            ('ifndef', r'^[ \t]*(`ifndef)[ \t]+(\w+)\b'),
            ('elsif', r'^[ \t]*(`elsif)[ \t]+(\w+)\b'),
            ('else', r'^[ \t]*(`else)\b'),
            ('endif', r'^[ \t]*(`endif)\b'),

            # Define
            ('define', r'^[ \t]*(`define)[ \t]+(\w+)(\()?'),

            # Undefine
            ('undef', r'^[ \t]*(`undef)[ \t]+(\w+)\b'),

            # Line directive (ignored)
            ('line', r'^[ \t]*(`line)[ \t]+\d+[ \t]+"(\\"|\\\\|[^\\"\n])*"[ \t]+\d+\b'),

            # Macro expansion
            # Check for parentheses in parameterized macros is handled later in
            # the arg scanner.
            ('macro', r'`(\w+)'),
        ]
        query_regex = re.compile(
            '|'.join('(?P<%s>%s)' % pair for pair in queries),
            re.DOTALL | re.MULTILINE
        )

        while True:
            m = query_regex.search(self._text, self._scan_idx)
            if not m:
                # Reached EOF
                break

            assert m.lastindex is not None

            # Process directives.
            # process_*() functions will advance the index as appropriate
            check_trailing = False
            if m.lastgroup == "ifdef":
                self.process_ifdef(m)
                check_trailing = True
            elif m.lastgroup == "ifndef":
                self.process_ifndef(m)
                check_trailing = True
            elif m.lastgroup == "elsif":
                self.process_elsif(m)
                check_trailing = True
            elif m.lastgroup == "else":
                self.process_else(m)
                check_trailing = True
            elif m.lastgroup == "endif":
                self.process_endif(m)
                check_trailing = True
            elif m.lastgroup == "macro":
                self.process_macro(m)
            elif m.lastgroup == "define":
                self.process_define(m)
            elif m.lastgroup == "undef":
                self.process_undef(m)
                check_trailing = True
            elif m.lastgroup == "line":
                # Discard this directive
                self.emit_segment(m.start(m.lastindex + 1))
                self._scan_idx = m.end()
                self._seg_start_idx = self._scan_idx
                check_trailing = True
            else:
                self._scan_idx = m.end()

            if check_trailing:
                # Check for trailing text on a macro directive line
                start, end = get_illegal_trailing_text_pos(self._text, self._scan_idx)
                if start is not None and end is not None:
                    self.env.msg.fatal(
                        "Unexpected text after preprocessor directive",
                        self.get_err_src_ref(start, end)
                    )



        # emit any trailing text
        self.emit_segment(len(self._text))

    def process_ifdef(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        identifier = m.group(m.lastindex + 2)

        if not self._conditional.is_idle:
            # push conditional context
            self._conditional_stack.append(self._conditional)
            self._conditional = ConditionalState(self._conditional.is_active)

        self._conditional.do_ifdef(identifier in self._macro_defs)
        self._scan_idx = m.end()
        self._seg_start_idx = self._scan_idx

    def process_ifndef(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        identifier = m.group(m.lastindex + 2)

        if not self._conditional.is_idle:
            # push conditional context
            self._conditional_stack.append(self._conditional)
            self._conditional = ConditionalState(self._conditional.is_active)

        self._conditional.do_ifndef(identifier in self._macro_defs)
        self._scan_idx = m.end()
        self._seg_start_idx = self._scan_idx

    def process_elsif(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        identifier = m.group(m.lastindex + 2)

        if not self._conditional.is_in_if_block or self._conditional.is_in_else_block:
            self.env.msg.fatal(
                "Unexpected `elsif",
                self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
            )

        self._conditional.do_elsif(identifier in self._macro_defs)
        self._scan_idx = m.end()
        self._seg_start_idx = self._scan_idx

    def process_else(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        if not self._conditional.is_in_if_block or self._conditional.is_in_else_block:
            self.env.msg.fatal(
                "Unexpected `else",
                self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
            )
        self._conditional.do_else()
        self._scan_idx = m.end()
        self._seg_start_idx = self._scan_idx

    def process_endif(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        if self._conditional.is_idle and not self._conditional_stack:
            self.env.msg.fatal(
                "Unexpected `endif",
                self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
            )

        self._conditional.do_endif()
        self._scan_idx = m.end()
        self._seg_start_idx = self._scan_idx

        if self._conditional_stack:
            # Pop conditional context
            self._conditional = self._conditional_stack.pop()

    def process_define(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        identifier = m.group(m.lastindex + 2)

        reserved_macro_names = (
            "include", "ifdef", "ifndef", "elsif", "else", "endif", "define",
            "undef", "line", "__LINE__", "__FILE__"
        )
        if identifier in reserved_macro_names:
            self.env.msg.fatal(
                "Macro name '%s' is a reserved keyword" % identifier,
                self.get_err_src_ref(m.start(m.lastindex + 2), m.end(m.lastindex + 2) - 1)
            )

        self._scan_idx = m.end(m.lastindex + 2)
        has_args = bool(m.group(m.lastindex + 3))

        if has_args:
            args = self.define_arg_scanner()
        else:
            args = []

        contents = self.define_contents_scanner()

        # Create macro defintion object
        macro = Macro(contents, args)
        if self._conditional.is_active:
            self._macro_defs[identifier] = macro

        self._seg_start_idx = self._scan_idx

    def process_undef(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start(m.lastindex + 1))

        if self._conditional.is_active:
            identifier = m.group(m.lastindex + 2)
            self._macro_defs.pop(identifier, None)

        self._scan_idx = m.end()
        self._seg_start_idx = self._scan_idx

    def process_macro(self, m: Match) -> None:
        assert m.lastindex is not None

        self.emit_segment(m.start())
        identifier = m.group(m.lastindex + 1)
        self._scan_idx = m.end()

        if not self._conditional.is_active:
            return

        macro_start_idx = m.start()

        if self._src_ref_override:
            macro_src_ref = self._src_ref_override
        else:
            assert isinstance(self._src_seg_map, segment_map.SegmentMap)
            macro_src_ref = SegmentedSourceRef(self._src_seg_map, m.start(), m.end() - 1)

        # Check if macro identifier is not one of the reserved directives
        # Preprocessor can end up here if the user did not provide the expected
        # args to a directive. The parser will instead fall back to thinking it
        # is a macro expansion
        reserved_macro_names = (
            "include", "ifdef", "ifndef", "elsif", "define",
            "undef", "line"
        )
        if identifier in reserved_macro_names:
            self.env.msg.fatal(
                "Preprocessor directive '`%s' is incomplete" % identifier,
                self.get_err_src_ref(m.start(), m.end() - 1)
            )

        # Lookup macro identifier
        if identifier not in self._macro_defs:
            self.env.msg.fatal(
                "Macro '`%s' has not been defined" % identifier,
                self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
            )
        macro = self._macro_defs[identifier]

        # Scan for macro args if necessary
        if macro.args:
            # scan for args
            raw_argv = self.macro_arg_scanner()

            # run each argv through the main scanner
            argv = []
            if self._conditional.is_active:
                for arg_text, arg_src_ref in raw_argv:
                    vpp = VerilogPreprocessor(self.env, arg_text, src_ref_override=arg_src_ref)
                    vpp._macro_defs = self._macro_defs
                    vpp._active_macro_stack = self._active_macro_stack
                    argv.append(vpp.preprocess()[0])
        else:
            argv = []

        macro_end_idx = self._scan_idx - 1

        # Push current macro into active stack
        if identifier in self._active_macro_stack:
            self.env.msg.fatal(
                "Found recursive macro expansion when processing '`%s'" % identifier,
                self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
            )
        self._active_macro_stack.append(identifier)

        # Emit macro text
        text = macro.render_macro(self, argv, macro_src_ref)
        self._output_text_segments.append(text)

        # If this is the top-level preprocessing pass (and therefore it
        # has a src seg map), also create a source tracking segment.
        if self._src_seg_map:
            text_len = len(text)
            segment = segment_map.MacroSegment(
                self._current_output_idx,
                self._current_output_idx + text_len - 1,
                macro_start_idx, macro_end_idx,
                self._src_seg_map
            )
            assert isinstance(self._output_seg_map, segment_map.SegmentMap)
            self._output_seg_map.segments.append(segment)
            self._current_output_idx += text_len

        # Done processing this macro
        self._active_macro_stack.pop()

        self._seg_start_idx = self._scan_idx

    def emit_segment(self, end_idx: int) -> None:
        if self._conditional.is_active:
            if self._seg_start_idx < end_idx:
                # Emit text segment
                seg = self._text[self._seg_start_idx:end_idx]
                self._output_text_segments.append(seg)

                # If this is the top-level preprocessing pass (and therefore it
                # has a src seg map), also create a source tracking segment.
                if self._src_seg_map:
                    segment = segment_map.UnalteredSegment(
                        self._current_output_idx,
                        self._current_output_idx + end_idx - self._seg_start_idx - 1,
                        self._seg_start_idx,
                        end_idx - 1,
                        self._src_seg_map
                    )
                    assert isinstance(self._output_seg_map, segment_map.SegmentMap)
                    self._output_seg_map.segments.append(segment)
                    # Advance the output offset
                    self._current_output_idx += end_idx - self._seg_start_idx

    #---------------------------------------------------------------------------
    # Define scanner
    #---------------------------------------------------------------------------
    def define_arg_scanner(self) -> List[str]:
        """
        Scan in define args starting from the current scan_idx.
        scan_idx is advanced to after the arg list.

        Returns a list of identifier strings
        """
        # Only implementing simple args. Not bothering with defaults for now.

        # Capture contents of parentheses
        arg_regex = re.compile(r'\(([ \t\w,]+)\)')
        m = arg_regex.match(self._text, self._scan_idx)
        if not m:
            self.env.msg.fatal(
                "Syntax error when parsing define args",
                self.get_err_src_ref(self._scan_idx, self._scan_idx)
            )
        current_arg_idx = self._scan_idx + 1
        self._scan_idx = m.end()

        # Extract individual arg identifiers
        raw_args = m.group(1).split(",")
        args = []
        for raw_arg in raw_args:
            m = re.fullmatch(r'\s*(\w+)\s*', raw_arg)
            if not m:
                self.env.msg.fatal(
                    "Syntax error when parsing define args",
                    self.get_err_src_ref(current_arg_idx, current_arg_idx + len(raw_arg) - 1)
                )
            args.append(m.group(1))
            current_arg_idx += len(raw_arg) + 1
        return args

    def define_contents_scanner(self) -> str:
        start_idx = self._scan_idx

        # First, do a rough scan of the text to determine the max extent of the
        # define, solely based on escaped newlines.
        # Find the first newline without a backslash
        unescaped_newline_regex = re.compile(r'[^\\\r]\r?$', re.MULTILINE)
        m = unescaped_newline_regex.search(
            self._text,
            # Intentionally rewind the scan index by 1 character in case it is
            # currently pointing at the terminating newline
            self._scan_idx - 1
        )
        if m:
            max_idx = m.end()
        else:
            # All lines up to EOF have a '\'.
            max_idx = len(self._text)

        # Next, re-scan to find if macro block actually ends earlier due to line
        # or block comment edge-cases. Need to properly hop over string
        # literals in case they contain comment-like characters.
        queries = [
            # Skip over macro-specific quote escapes
            ('esc', r'(?:`"|`\\`")'),
            # Skip over string literals
            ('str', r'"(\\"|\\\\|\\\r?\n|[^\\"])*"'),
            # Skip over block comments that are contained within the same line
            ('slbc', r'/\*[^\n]*?\*/'),

            # Define ends on a line comment, or the start of a multi-line block comment
            ('end1', r'//'),
            ('end2', r'/\*'),
        ]
        query_regex = re.compile(
            '|'.join('(?P<%s>%s)' % pair for pair in queries),
            re.DOTALL
        )
        while True:
            m = query_regex.search(self._text, self._scan_idx, max_idx)
            if not m:
                # Reached end of macro range without running into a comment
                self._scan_idx = max_idx
                break

            if m.lastgroup in {"end1", "end2"}:
                # Reached end of define due to a comment
                self._scan_idx = m.start()
                break

            # Advance index for next iteration
            self._scan_idx = m.end()

        # Strip any newline escapes
        return re.sub(r'\\(\r?\n)', r'\1', self._text[start_idx:self._scan_idx]).strip()

    #---------------------------------------------------------------------------
    # Macro Arg Scanner
    #---------------------------------------------------------------------------
    def macro_arg_scanner(self) -> List[Tuple[str, SourceRefBase]]:
        """
        When a macro is instantiated and has args, this scanner parses them
        and returns the raw extracted arg text

        Returns a list of tuples:
            (arg_text, SourceRefBase)
        """

        # First, skip any whitespace until the open parentheses are reached
        m = re.compile(r'\s*\(').match(self._text, self._scan_idx)
        if not m:
            self.env.msg.fatal(
                "Expected arguments to macro. Got none.",
                self.get_err_src_ref(self._scan_idx, self._scan_idx)
            )
        self._scan_idx = m.end()
        args_start_idx = self._scan_idx - 1

        queries = [
            # Skip over stuff
            ('mlc', r'/\*.*?\*/'),
            ('slc', r'//.*?$'),
            ('str', r'"(\\"|\\\\|[^\\"])*"'),

            # Detect punctuation
            ('punc', r'([\[\]\{\}\(\),])'),
        ]
        query_regex = re.compile(
            '|'.join('(?P<%s>%s)' % pair for pair in queries),
            re.DOTALL | re.MULTILINE
        )

        enclosures_stack = [] # type: List[str]
        argvs = []
        argv_start_idx = self._scan_idx
        punc_pairs = {
            '[':']',
            '{':'}',
            '(':')',
        }

        for m in query_regex.finditer(self._text, self._scan_idx):
            assert m.lastindex is not None

            if m.lastgroup != "punc":
                continue

            punc = m.group(m.lastindex + 1)
            if not enclosures_stack and punc == ',':
                # macro arg separator
                argv = self._text[argv_start_idx:m.start()].strip()
                if self._src_ref_override:
                    argvs.append((argv, self._src_ref_override))
                else:
                    assert isinstance(self._src_seg_map, segment_map.SegmentMap)
                    src_ref = SegmentedSourceRef(self._src_seg_map, argv_start_idx, m.start() - 1)
                    argvs.append((argv, src_ref))
                argv_start_idx = m.end()
            elif not enclosures_stack and punc == ')':
                # Final closing parentheses. End of args!
                argv = self._text[argv_start_idx:m.start()].strip()
                if self._src_ref_override:
                    argvs.append((argv, self._src_ref_override))
                else:
                    assert isinstance(self._src_seg_map, segment_map.SegmentMap)
                    src_ref = SegmentedSourceRef(self._src_seg_map, argv_start_idx, m.start() - 1)
                    argvs.append((argv, src_ref))
                self._scan_idx = m.end()
                break
            elif punc in {'[', '{', '('}:
                enclosures_stack.append(punc)
            elif punc in {']', '}', ')'}:
                if not enclosures_stack:
                    self.env.msg.fatal(
                        "Unexpected '%s' while parsing macro arguments." % punc,
                        self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
                    )
                c = enclosures_stack.pop()
                if punc != punc_pairs[c]:
                    self.env.msg.fatal(
                        "Unexpected '%s' while parsing macro arguments." % punc,
                        self.get_err_src_ref(m.start(m.lastindex + 1), m.end(m.lastindex + 1) - 1)
                    )
        else:
            self.env.msg.fatal(
                "Reached end of text before all macro args could be parsed",
                self.get_err_src_ref(args_start_idx, args_start_idx)
            )

        return argvs

def get_illegal_trailing_text_pos(text: str, idx: int) -> Tuple[Optional[int], Optional[int]]:
    """
    Scan the remainder of a line for illegal text.
    Verilog preprocessor directives require that there be no trailing text,
    except for comments.

    If illegal text is found, return the start/end index of it as a tuple.
    Otherwise return (None, None)
    """

    # First, scan ahead past all block comments
    bc_regex = re.compile(r'(?:[ \t]*/\*.*?\*/)*[ \t]*')
    m = bc_regex.match(text, idx)
    assert m is not None
    idx = m.end()

    # Assert that any remaining text is a comment
    rem_regex = re.compile(r'(:?//.*|/\*.*)?\r?$', re.MULTILINE)
    m = rem_regex.match(text, idx)
    if m:
        return (None, None)
    else:
        # Did not match. Find end of the line
        eol_regex = re.compile(r'.+$', re.MULTILINE)
        m = eol_regex.match(text, idx)
        assert m is not None
        return (m.start(), m.end() - 1)


class ConditionalState:
    def __init__(self, parent_is_active: bool=True):
        self.is_in_if_block = False
        self.is_in_else_block = False

        # Is in a block that will be emitted
        self.is_active = parent_is_active

        # Conditional has been exhausted
        self.is_done = False

        # Parent conditional is in an active state
        self.parent_is_active = parent_is_active

    @property
    def is_idle(self) -> bool:
        return not (self.is_in_else_block or self.is_in_if_block)

    def do_ifdef(self, result: bool) -> None:
        assert self.is_idle
        self.is_in_if_block = True
        self.is_in_else_block = False
        self.is_done = False
        if result:
            self.is_active = self.parent_is_active
        else:
            self.is_active = False

    def do_ifndef(self, result: bool) -> None:
        self.do_ifdef(not result)

    def do_elsif(self, result: bool) -> None:
        assert self.is_in_if_block
        assert not self.is_in_else_block

        if self.is_done:
            return

        if self.is_active:
            # prev block was active
            self.is_active = False
            self.is_done = True
        elif result:
            # success!
            self.is_active = self.parent_is_active

    def do_else(self) -> None:
        assert self.is_in_if_block
        assert not self.is_in_else_block

        self.is_in_if_block = False
        self.is_in_else_block = True

        if self.is_done:
            return

        if self.is_active:
            # prev block was active
            self.is_active = False
            self.is_done = True
        else:
            self.is_active = self.parent_is_active

    def do_endif(self) -> None:
        assert not self.is_idle
        self.is_in_if_block = False
        self.is_in_else_block = False
        self.is_active = self.parent_is_active
        self.is_done = False


class Macro:
    def __init__(self, contents: str, args: List[str]):
        self.args = args
        self.segments = self.prepare_segments(contents)

    def prepare_segments(self, contents: str) -> List[Union[int, str]]:
        """
        Prepares the macro contents:
        - remove macro-specific quote escape sequences
        - remove token paste ``
        - remove comments
        - capture arg tokens

        results in a list of segments:
            If an entry is a string, it gets emitted as-is
            If an entry is an integer, it represents the value of the arg it indexes
        """

        queries = [
            ('esc1', r'`"'),
            ('esc2', r'`\\`"'),
            ('tp', r'``'),
            ('str', r'"(\\"|\\\\|[^\\"])*"'),
            ('slbc', r'/\*.*?\*/'),
        ]
        if self.args:
            queries.append(
                ('arg', r'\b(%s)\b' % '|'.join(self.args))
            )

        query_regex = re.compile(
            '|'.join('(?P<%s>%s)' % pair for pair in queries)
        )

        seg_start_idx = 0
        segments = [] # type: List[Union[int, str]]
        for m in query_regex.finditer(contents):
            assert m.lastindex is not None

            if m.lastgroup == "esc1":
                # replace with unescaped version
                segments.append(contents[seg_start_idx:m.start()])
                seg_start_idx = m.end()
                segments.append('"')
            elif m.lastgroup == "esc2":
                # replace with unescaped version
                segments.append(contents[seg_start_idx:m.start()])
                seg_start_idx = m.end()
                segments.append('\\"')
            elif m.lastgroup in ("tp", "slbc"):
                # discard
                segments.append(contents[seg_start_idx:m.start()])
                seg_start_idx = m.end()
            elif m.lastgroup == "arg":
                # replace with arg index
                segments.append(contents[seg_start_idx:m.start()])
                seg_start_idx = m.end()

                arg_name = m.group(m.lastindex + 1)
                arg_idx = self.args.index(arg_name)
                segments.append(arg_idx)

        segments.append(contents[seg_start_idx:])
        return segments

    def render_macro(self, parent_vpp: VerilogPreprocessor, argv: List[str], src_ref: SourceRefBase) -> str:
        if len(argv) != len(self.args):
            parent_vpp.env.msg.fatal(
                "Macro expansion requires %d arguments. Got %d instead"
                % (len(self.args), len(argv)),
                src_ref
            )

        # Join text segments. Replace arg values along the way
        text_segments = []
        for seg in self.segments:
            if isinstance(seg, int):
                text_segments.append(argv[seg])
            else:
                text_segments.append(seg)
        text = "".join(text_segments)

        # Perform substitutions on macro text before emitting it
        # Copy macro context
        # Intentionally do not pass the conditional state to effectively reset it
        vpp = VerilogPreprocessor(parent_vpp.env, text, src_ref_override=src_ref)
        vpp._macro_defs = parent_vpp._macro_defs
        vpp._active_macro_stack = parent_vpp._active_macro_stack
        text, _ = vpp.preprocess()

        return text



from typing import Optional

from .. import rdltypes
from .. import component as comp
from .. import node as m_node

class PropertyValueReference(rdltypes.PropertyReference):
    """
    Directly references the value that was assigned to said property.
    """
    def _validate(self) -> None:
        # validate that this prop is set in the target (is not None)
        target_value = self.node.get_property(self.name)
        if target_value is None:
            self.env.msg.error(
                "Assignment references the value of property '%s', but its value was never set for instance '%s'"
                % (self.name, self.node.inst_name),
                self.src_ref
            )

        # Property value references could theoretically create a circular loop
        # validate that circular references do not exist
        ref_value = target_value
        while isinstance(ref_value, PropertyValueReference):
            if ref_value == self:
                # Looped back to a property that points to self.
                # Each reference in the chain will also emit its own error
                self.env.msg.error(
                    "Assignment creates a circular reference",
                    self.src_ref
                )
                break
            ref_value = ref_value.node.get_property(ref_value.name)


class PropertyValueReferenceFieldwidth(PropertyValueReference):
    """
    Similar to PropertyValueReference, but is limited to fields and tracks
    the field's width
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        assert isinstance(self.node, m_node.FieldNode)
        return self.node.width


class RealOrInferredVectorReference(PropertyValueReference):
    """
    References the vector that was directly assign to the property, or inferred
    by setting the property to True
    """
    complementary_prop = None # type: Optional[str]
    def _validate(self) -> None:
        super()._validate()
        # validate that this property is enabled in the target (is not False)
        # If complementary_prop is defined, check the partner property too before failing
        target_value = self.node.get_property(self.name)
        if target_value is False and self.complementary_prop is not None:
            # try its complement
            target_value = self.node.get_property(self.complementary_prop)
        if target_value is False:
            self.env.msg.error(
                "Assignment references property '%s', but the signal it represents was never defined or enabled for instance '%s'"
                % (self.name, self.node.inst_name),
                self.src_ref
            )

#-------------------------------------------------------------------------------
# Reductions
#-------------------------------------------------------------------------------
class ReductionPropRef(rdltypes.PropertyReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_anded(ReductionPropRef):
    pass

class PropRef_ored(ReductionPropRef):
    pass

class PropRef_xored(ReductionPropRef):
    pass

#-------------------------------------------------------------------------------
# Counter
#-------------------------------------------------------------------------------
class CounterPropRef(rdltypes.PropertyReference):
    def _validate(self) -> None:
        if not self.node.get_property('counter'):
            self.env.msg.error(
                "Reference to property '%s' is illegal because '%s' is not a counter"
                % (self.name, self.node.inst_name),
                self.src_ref
            )

class CounterThresholdPropRef(CounterPropRef):
    def _validate(self) -> None:
        super()._validate()
        # validate that the counter actually sets a threshold
        target_value = self.node.get_property(self.name)
        if target_value is False:
            self.env.msg.error(
                "Reference to property '%s' is illegal because the target field does not define any thresholds"
                % self.name,
                self.src_ref
            )

class PropRef_incr(CounterPropRef):
    """
    References the increment event signal
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_incrsaturate(CounterPropRef):
    """
    referencing the counter’s saturate output, which is a single bit value
    indicating whether the saturation has occurred
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_saturate(PropRef_incrsaturate):
    """
    alias of incrsaturate.
    """

class PropRef_incrthreshold(CounterThresholdPropRef):
    """
    Referencing the counter’s threshold output, which is a single bit value
    indicating whether the threshold has been crossed
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_threshold(PropRef_incrthreshold):
    """
    alias of incrthreshold
    """

class PropRef_incrvalue(PropertyValueReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        # Does not correspond to the field's width.
        # not always knowable
        return None

class PropRef_decr(CounterPropRef):
    """
    References the decrement event signal
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_decrsaturate(CounterPropRef):
    """
    referencing the counter’s saturate output, which is a single bit value
    indicating whether the saturation has occurred
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_decrthreshold(CounterThresholdPropRef):
    """
    Referencing the counter’s threshold output, which is a single bit value
    indicating whether the threshold has been crossed
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_decrvalue(PropertyValueReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        # Does not correspond to the field's width.
        # not always knowable
        return None

class PropRef_overflow(CounterPropRef):
    """
    asserted when counter overflows or wraps
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

    def _validate(self) -> None:
        super()._validate()

        if self.node.get_property('incrsaturate') is not False:
            self.env.msg.error(
                "Reference to property '%s' is illegal because the target field will never overflow"
                % self.name,
                self.src_ref
            )


class PropRef_underflow(CounterPropRef):
    """
    asserted when counter underflows or wraps.
    """
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

    def _validate(self) -> None:
        super()._validate()

        if self.node.get_property('decrsaturate') is not False:
            self.env.msg.error(
                "Reference to property '%s' is illegal because the target field will never underflow"
                % self.name,
                self.src_ref
            )

#-------------------------------------------------------------------------------
# Access
#-------------------------------------------------------------------------------
class PropRef_swacc(rdltypes.PropertyReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_swmod(rdltypes.PropertyReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_swwe(RealOrInferredVectorReference):
    allowed_inst_type = comp.Field
    complementary_prop = "swwel"

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_swwel(RealOrInferredVectorReference):
    allowed_inst_type = comp.Field
    complementary_prop = "swwe"

    @property
    def width(self) -> Optional[int]:
        return 1

#-------------------------------------------------------------------------------
# HW Signals
#-------------------------------------------------------------------------------
class PropRef_we(RealOrInferredVectorReference):
    allowed_inst_type = comp.Field
    complementary_prop = "wel"

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_wel(RealOrInferredVectorReference):
    allowed_inst_type = comp.Field
    complementary_prop = "we"

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_hwset(RealOrInferredVectorReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_hwclr(RealOrInferredVectorReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1

#-------------------------------------------------------------------------------
# Interrupts
#-------------------------------------------------------------------------------
class PropRef_intr(rdltypes.PropertyReference):
    """
    Represents the inclusive OR of all the interrupt bits in a register after
    any field enable and/or field mask logic has been applied.
    """
    allowed_inst_type = comp.Reg
    def _validate(self) -> None:
        # validate reg contains at least one field that is intr
        for field in self.node.fields():
            if field.get_property('intr'):
                break
        else:
            self.env.msg.error(
                "'intr' property reference is illegal because target register does not contain any interrupt fields",
                self.src_ref
            )

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_halt(rdltypes.PropertyReference):
    """
    Represents the inclusive OR of all the interrupt bits in a register after
    any field haltenable and/or field haltmask logic has been applied.
    """
    allowed_inst_type = comp.Reg
    def _validate(self) -> None:
        # 10.8.1-c: shall only be present if haltmask or haltenable is
        # specified on at least one field in the register.
        for field in self.node.fields():
            if field.get_property('haltenable') or field.get_property('haltmask'):
                break
        else:
            self.env.msg.error(
                "'halt' property reference is illegal because target register does not contain any fields with 'haltenable' or 'haltmask' set.",
                self.src_ref
            )

    @property
    def width(self) -> Optional[int]:
        return 1

class PropRef_haltenable(PropertyValueReferenceFieldwidth):
    pass

class PropRef_haltmask(PropertyValueReferenceFieldwidth):
    pass

class PropRef_enable(PropertyValueReferenceFieldwidth):
    pass

class PropRef_mask(PropertyValueReferenceFieldwidth):
    pass

#-------------------------------------------------------------------------------
class PropRef_hwenable(PropertyValueReferenceFieldwidth):
    pass

class PropRef_hwmask(PropertyValueReferenceFieldwidth):
    pass

class PropRef_next(PropertyValueReferenceFieldwidth):
    pass

class PropRef_reset(PropertyValueReferenceFieldwidth):
    pass

class PropRef_resetsignal(PropertyValueReference):
    allowed_inst_type = comp.Field

    @property
    def width(self) -> Optional[int]:
        return 1
